#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "game.h"

#include "qmap.h"

#include "color.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "tex.h"
#include "tex_all.h"


static int SCS_TexSort(const char *t1,const char *t2)
{
   return stricmp(t1,t2);
}


#define TEX_SIZE 32

static void CreateTex(texture_t *t,const char *name,int c1,int c2)
{
   int i;
   unsigned char *c;

   t->data=Q_malloc(TEX_SIZE*TEX_SIZE);
   t->color=-1;
   if (!t->data)
   {
      Abort("SCS_CreateTex","Out of memory!");
   }
   t->dsx = t->rsx = TEX_SIZE;
   t->dsy = t->rsy = TEX_SIZE;

   for (i=TEX_SIZE*TEX_SIZE,c=t->data;i;i--,c++)
      if (i&1) *c=c1; else *c=c2;

   strcpy(t->name,name);
}

static int SCS_ReadCache(int verbose)
{
   int i;
   int r,g,b;
   texture_t *t;

   ClearCache();

	M.num_textures=4*4*4+4;

	M.Cache=(texture_t *)Q_malloc(sizeof(texture_t)*M.num_textures);
   if (!M.Cache)
   {
      if (verbose)
         HandleError("SCS_ReadCache","Out of memory!");
      return 0;
   }

   for (i=0,t=M.Cache;i<4*4*4;i++,t++)
   {
      t->data=Q_malloc(TEX_SIZE*TEX_SIZE);
      t->color=-1;
      if (!t->data)
      {
         Q_free(M.Cache);
         M.Cache=0;
         M.num_textures=0;
         return 0;
      }
      t->dsx = t->rsx = TEX_SIZE;
      t->dsy = t->rsy = TEX_SIZE;

      r=( i    &3)*85;
      g=((i>>2)&3)*85;
      b=((i>>4)&3)*85;

      sprintf(t->name,"c_%02x%02x%02x",r,g,b);
      memset(t->data,i,TEX_SIZE*TEX_SIZE);
   }

   CreateTex(t++,"ground",63,55);
   CreateTex(t++,"grd_drop",63,57);
   CreateTex(t++,"grd_ign",63,59);

   CreateTex(t++,"nodraw",62,48);

   return 1;
}


static game_t Game_SCS=
{
   "SCS",
   {
      0,

      SCS_TexSort,
      SCS_ReadCache,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
   },
   {
      QMap_Load,
      QMap_Save,
      QMap_SaveVisible,
      QMap_LoadGroup,
      QMap_SaveGroup,
      QMap_Profile
   },
   {
      L_QUAKE,
      0,
      1
   },
   {
      1
   },
   ".pts"
};


void SCS_Init(void);
void SCS_Init(void)
{
   Game=Game_SCS;
   qmap_loadtexinfo=QMap_LoadTexInfo;
   qmap_savetexinfo=QMap_SaveTexInfo;

   {
      int i;
      unsigned char tex_pal[768];
      unsigned char *tp;

      memset(tex_pal,0,sizeof(tex_pal));
      for (i=0,tp=tex_pal;i<64;i++)
      {
         *tp++=( i    &3)*21;
         *tp++=((i>>2)&3)*21;
         *tp++=((i>>4)&3)*21;
      }

      for (i=0;i<6*6*5;i++)
      {
         *tp++=((i   )%6)*12;
         *tp++=((i/ 6)%6)*12;
         *tp++=((i/36)%5)*15;
      }
      SetTexturePal(tex_pal);
   }
}


#include "brush.h"
#include "entity.h"


#define MAXV 128
typedef struct
{
   int nverts;
   int v[MAXV];
   char tex[64];
} face_t;


static face_t *faces;
static int nfaces;

static vec3_t *verts;
static int nverts;


static void ClearModel(void)
{
   if (faces) Q_free(faces);
   if (verts) Q_free(verts);
   nverts=nfaces=0;
   verts=NULL;
   faces=NULL;
}


static int AddVert(float x,float y,float z)
{
   int i;
   vec3_t *v;
   for (i=0,v=verts;i<nverts;i++,v++)
      if (v->x==x && v->y==y && v->z==z)
         return i;

   verts=Q_realloc(verts,sizeof(vec3_t)*(nverts+1));
   if (!verts)
      Abort("SCS_AddVert","Out of memory!");
   verts[nverts].x=x;
   verts[nverts].y=y;
   verts[nverts].z=z;
   nverts++;
   return nverts-1;
}

static void AddBrush(brush_t *b,float dx,float dy,float dz)
{
   int i,j;
   plane_t *p;
   face_t *f;

   for (i=0,p=b->plane;i<b->num_planes;i++,p++)
   {
      if (!strcmp(p->tex.name,"nodraw"))
         continue;

      faces=Q_realloc(faces,sizeof(face_t)*(nfaces+1));
      if (!faces)
         Abort("SCS_AddBrush","Out of memory!");
      f=&faces[nfaces++];
   
      f->nverts=p->num_verts;
      for (j=0;j<p->num_verts;j++)
      {
         f->v[j]=AddVert(
            b->verts[p->verts[j]].x+dx,
            b->verts[p->verts[j]].y+dy,
            b->verts[p->verts[j]].z+dz);
      }
      strcpy(f->tex,p->tex.name);
   }
}


static void WriteModel(FILE *f)
{
   int i,j;
   vec3_t *v;
   face_t *fc;

   fprintf(f,"verts %i\n",nverts);
   for (i=0,v=verts;i<nverts;i++,v++)
   {
      fprintf(f,"%g %g %g\n",v->x,v->y,v->z);
   }
   fprintf(f,"faces %i\n",nfaces);
   for (i=0,fc=faces;i<nfaces;i++,fc++)
   {
      fprintf(f,"%i (",fc->nverts);
      for (j=0;j<fc->nverts;j++)
         fprintf(f," %i",fc->v[j]);
      fprintf(f," ) %s\n",fc->tex);
   }
}

static void AddGBrush(brush_t *b,FILE *f)
{
   plane_t *p;
   int i,j;
   float min,max,a;
   vec3_t edge;

   for (i=0,p=b->plane;i<b->num_planes;i++,p++)
   {
      if (!strcmp(p->tex.name,"grd_ign"))
         continue;
      if (p->normal.y)
         continue;

      if (!strcmp(p->tex.name,"ground"))
         fprintf(f,"ground ");
      else
         fprintf(f,"platform ");

      fprintf(f,"( %g %g %g )",
         p->normal.x,p->normal.z,p->dist);

      edge.x= p->normal.z;
      edge.y=0;
      edge.z=-p->normal.x;

      min=max=edge.x*b->verts[p->verts[0]].x+
              edge.z*b->verts[p->verts[0]].z;
      for (j=0;j<p->num_verts;j++)
      {
         a=edge.x*b->verts[p->verts[j]].x+
           edge.z*b->verts[p->verts[j]].z;
         if (a<min) min=a;
         if (a>max) max=a;
      }
      fprintf(f," ( %g %g %g %g )\n",
         edge.x,edge.z,min,max);
   }
}


void SCS_Output(void);
void SCS_Output(void)
{
   if (GetKeyValue(M.WorldSpawn,"back"))
   {
      brush_t *b;
      FILE *f;
      char name[256];
      int rc,gc,bc;

      sscanf(GetKeyValue(M.WorldSpawn,"back"),"%i %i %i",&rc,&gc,&bc);

      ClearModel();
      for (b=M.BrushHead;b;b=b->Next)
      {
         if (!strncmp(b->plane[0].tex.name,"gr",2))
            continue;
         AddBrush(b,0,0,0);
      }
      strcpy(name,M.mapfilename);
      *strrchr(name,'.')=0;
      strcat(name,".wmd");
      f=fopen(name,"wt");
      fprintf(f,"world %02x%02x%02x\n",rc,gc,bc);
      WriteModel(f);
      fclose(f);
      ClearModel();

      strcpy(name,M.mapfilename);
      *strrchr(name,'.')=0;
      strcat(name,".grd");
      f=fopen(name,"wt");
      for (b=M.BrushHead;b;b=b->Next)
      {
         if (strncmp(b->plane[0].tex.name,"gr",2))
            continue;
         AddGBrush(b,f);
      }
      fclose(f);
   }
   else
   {
      group_t *g;
      brush_t *b;
      int fn;
      FILE *f;
      float dx,dy,dz;

      char name[256];
      strcpy(name,M.mapfilename);
      *strrchr(name,'.')=0;
      strcat(name,".mdl");
      f=fopen(name,"wt");

      fprintf(f,"model\n");
      for (g=M.GroupHead;g;g=g->Next)
      {
         if (g->groupname[0]!='f')
            continue;
         if (sscanf(g->groupname,"f%i",&fn)!=1)
            continue;
         fprintf(f,"frame %i\n",fn);

         ClearModel();

         dx=dy=dz=0;
         for (b=M.BrushHead;b;b=b->Next)
         {
            if (b->Group==g && !strcmp(b->plane[0].tex.name,"ground"))
            {
               dx=-(b->center.x);
               dy=-(b->center.y);
               dz=-(b->center.z+8+32);
            }
         }

         for (b=M.BrushHead;b;b=b->Next)
         {
            if (b->Group!=g)
               continue;
            if (!strcmp(b->plane[0].tex.name,"ground"))
               continue;

            AddBrush(b,dx,dy,dz);
         }

         WriteModel(f);
         ClearModel();
      }
      fclose(f);
   }
}

