#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "csg.h"

#include "brush.h"
#include "game.h"
#include "geom.h"
#include "map.h"
#include "memory.h"


cface_t *CSG_FaceToCFace(cface_t *dst,const struct brush_s *b,int num)
{
   face_t *f;
   int i;

   if (!(b->bt->flags&BRUSH_F_NORMAL)) return NULL;
   if (!dst)
   {
      dst=Q_malloc(sizeof(cface_t));
      if (!dst) return NULL;
   }
   memset(dst,0,sizeof(cface_t));

   f=&b->faces[num];
   
   dst->p.n=f->normal;
   dst->p.dist=f->dist;
   dst->tex=f->tex;
   dst->tex.g=NULL;
   G_Texdef_Copy(&dst->tex,&f->tex);
   dst->ba=f->a;

   dst->num_pts=f->num_verts;
   if (dst->num_pts>MAX_FACE_POINTS) return NULL;
   for (i=0;i<f->num_verts;i++)
      dst->pts[i]=b->verts[f->verts[i]];
   return dst;
}

cbrush_t *CSG_BrushToCBrush(cbrush_t *dst,const struct brush_s *b)
{
   int i;
   
   if (!(b->bt->flags&BRUSH_F_NORMAL)) return NULL;
   if (!dst)
   {
      dst=Q_malloc(sizeof(cbrush_t));
      if (!dst) return NULL;
   }
   memset(dst,0,sizeof(cbrush_t));
   dst->num_faces=b->num_faces;
   dst->faces=Q_malloc(sizeof(cface_t)*b->num_faces);
   if (!dst->faces) return NULL;
   for (i=0;i<dst->num_faces;i++)
   {
      if (!CSG_FaceToCFace(&dst->faces[i],b,i))
         return NULL;
   }
   return dst;
}

void CSG_FreeFace(cface_t *f)
{
   if (f->tex.g)
      G_Texdef_Free(&f->tex);
}

void CSG_FreeBrush(cbrush_t *b)
{
   int i;
   for (i=0;i<b->num_faces;i++)
      if (b->faces[i].tex.g)
         G_Texdef_Free(&b->faces[i].tex);
   Q_free(b->faces);
   b->faces=NULL;
}


int CSG_CalcPlane(cface_t *f)
{
   vec3_t v1,v2,norm;

   v1.x = f->planepts[1].x - f->planepts[0].x;
   v1.y = f->planepts[1].y - f->planepts[0].y;
   v1.z = f->planepts[1].z - f->planepts[0].z;
   
   v2.x = f->planepts[1].x - f->planepts[2].x;
   v2.y = f->planepts[1].y - f->planepts[2].y;
   v2.z = f->planepts[1].z - f->planepts[2].z;
   
   CrossProd(v1, v2, &norm);
   Normalize(&norm);
   
   /* Convert to normal / dist plane */
   f->p.n.x = norm.x;
   f->p.n.y = norm.y;
   f->p.n.z = norm.z;
         
   f->p.dist = DotProd(norm,f->planepts[0]);

   if (DotProd(f->p.n,f->p.n)<0.9)
      return -1;
   else
      return 0;
}

void CSG_CalcPlanes(cbrush_t *b)
{
   int i;
   for (i=0;i<b->num_faces;i++)
      CSG_CalcPlane(&b->faces[i]);
}


void CSG_MakeBoxOnPlane(cface_t *f)
{
   int x;
   double max,v;
   vec3_t org,vright,vup;

   /* find the major axis*/
   max = f->p.n.x;
   x = 0;
   if (fabs(f->p.n.y)>max)
   {
      max=fabs(f->p.n.y);
      x=1;
   }
   if (fabs(f->p.n.z)>max)
      x=2;
   
   vup.x=vup.y=vup.z=0;
   switch (x)
   {
   case 0:
   case 1:
      vup.z=1;
      break;
   case 2:
      vup.x=1;
      break;
   }

   v = DotProd(vup,f->p.n);
   vup.x-=v*f->p.n.x;
   vup.y-=v*f->p.n.y;
   vup.z-=v*f->p.n.z;
   Normalize(&vup);

   org.x=f->p.n.x*f->p.dist;
   org.y=f->p.n.y*f->p.dist;
   org.z=f->p.n.z*f->p.dist;

   CrossProd(vup,f->p.n,&vright);

      vup.x*=8192;    vup.y*=8192;    vup.z*=8192;
   vright.x*=8192; vright.y*=8192; vright.z*=8192;

   /* project a really big axis aligned box onto the plane*/
   f->num_pts = 4;

   VEC_Sub(f->pts[0],org,vright);
   VEC_Add(f->pts[0],f->pts[0],vup);
   
   VEC_Add(f->pts[1],org,vright);
   VEC_Add(f->pts[1],f->pts[1],vup);
   
   VEC_Add(f->pts[2],org,vright);
   VEC_Sub(f->pts[2],f->pts[2],vup);
   
   VEC_Sub(f->pts[3],org,vright);
   VEC_Sub(f->pts[3],f->pts[3],vup);
}


int CSG_MakeFaces(cbrush_t *b)
{
   cface_t Face1;
   cface_t NewFace;
   cface_t *f;
   plane_t p;
   int i,j,k,l;
   int flag;
   
   for (i=0;i<b->num_faces;i++) b->faces[i].misc=0;

   for (i=0;i<b->num_faces;i++)
   {
      if (b->faces[i].misc==2) continue;
         
      NewFace = b->faces[i];
      CSG_MakeBoxOnPlane(&NewFace);

      flag=0;
      for (j=0;j<b->num_faces;j++)
      {
         if (j==i) continue;

         VEC_FlipPlane(p,b->faces[j].p);
         
         k = CSG_SplitFace(&NewFace,p,&Face1,NULL,0);
         if (k==-4)
         {
/*          printf("Duplicate plane encountered, face removed\n");*/
            b->faces[j].misc=2;
            continue;
         }
         if (k==-3)
         {
/*          printf("Error splitting face, face removed\n");*/
            flag=1;
            break;
         }
         if (k==-1)
         {
/*          printf("Face exceeded brush bounds, face removed\n");*/
            flag=1;
            break;
         }
         NewFace = Face1;
      }
      if (flag)
      {
         b->faces[i].misc = flag;
         continue;
      }
      
      b->faces[i] = NewFace;
      b->faces[i].misc = 0;

      for (j=0;j<b->faces[i].num_pts;j++)
         SnapVecToInt(&b->faces[i].pts[j]);
   }

/*
Extra tests to get rid of bad faces. First, join vertices that are close
together. Then, remove any degenerated faces. Finally, check that the
face has a normal. If it doesn't, remove it.
*/

   for (i=0;i<b->num_faces;i++)
   {
      f=&b->faces[i];
      if (f->misc)
         continue;

      for (j=0;j<f->num_pts;)
      {
         k=(j+1)%f->num_pts;
         if ( (fabs(f->pts[j].x - f->pts[k].x)<0.1) &&
              (fabs(f->pts[j].y - f->pts[k].y)<0.1) &&
              (fabs(f->pts[j].z - f->pts[k].z)<0.1))
         {
            f->num_pts--;
            if (k)
               for (l=k;l<f->num_pts;l++)
                  f->pts[l]=f->pts[l+1];
         }
         else
            j++;
      }
      if (f->num_pts<3)
         f->misc=1;
   }

   j=0;
   for (i=0;i<b->num_faces;i++)
   {
      if (b->faces[i].misc)
      {
         CSG_FreeFace(&b->faces[i]);
         j++;
      }
      else
      if (j)
         b->faces[i-j]=b->faces[i];
   }
   b->num_faces-=j;

   if (b->num_faces<4)
   {
      Q_free(b->faces);
      return 1;
   }
   {
      cface_t *f;
      f=Q_realloc(b->faces,sizeof(cface_t)*b->num_faces);
      if (f) b->faces=f;
   }

   return 0;
}


struct brush_s *CSG_BuildBrush(struct brush_s *dst,cbrush_t *b)
{
   edge_t tmpedge;
   int    i,j,k,l;

   if (CSG_MakeFaces(b))
      return NULL;

   if (!dst)
   {
      dst=B_New(BRUSH_NORMAL);
      if (!dst)
      {
         Q_free(b->faces);
         return NULL;
      }
   }

   dst->faces = (face_t *)Q_malloc(sizeof(face_t) * b->num_faces);
   if (dst->faces==NULL)
   {
      Q_free(b->faces);
      return NULL;
   }
   memset(dst->faces,0,sizeof(face_t)*b->num_faces);
   dst->num_faces = b->num_faces;
   
   dst->edges = NULL;
   dst->num_edges = 0;
   dst->verts = NULL;
   dst->num_verts = 0;
   
   for (i=0;i<b->num_faces;i++)
   {
      dst->faces[i].tex = b->faces[i].tex;
      dst->faces[i].a   = b->faces[i].ba;
      
      dst->faces[i].normal = b->faces[i].p.n;
      dst->faces[i].dist   = b->faces[i].p.dist;

      dst->faces[i].verts = (int *)Q_malloc(sizeof(int) * (b->faces[i].num_pts));
      if (!dst->faces[i].verts)
      {
         Q_free(b->faces);
         return NULL;
      }

      dst->faces[i].num_verts = b->faces[i].num_pts;
      
      for (j=0;j<b->faces[i].num_pts;j++)
      {
         for (k=0;k<dst->num_verts;k++)
            if (VertsAreEqual(dst->verts[k],b->faces[i].pts[j]))
            {
               dst->faces[i].verts[j]=k;
               break;
            }

         if (k==dst->num_verts)
         {
            dst->verts = (vec3_t *)Q_realloc(dst->verts,sizeof(vec3_t) * (dst->num_verts+1));
            if (!dst->verts)
            {
               Q_free(b->faces);
               return NULL;
            }
            dst->verts[dst->num_verts] = b->faces[i].pts[j];
            dst->faces[i].verts[j] = dst->num_verts;
            dst->num_verts++;
         }
      }

      for (j=0;j<dst->faces[i].num_verts;j++)
      {
         if (j==dst->faces[i].num_verts-1)
            l=0;
         else
            l=j+1;
            
         tmpedge.startvertex = dst->faces[i].verts[j];
         tmpedge.endvertex   = dst->faces[i].verts[l];
         for (k=0;k<dst->num_edges;k++)
            if (EdgesAreEqual(dst->edges[k],tmpedge))
               break;
               
         if (k==dst->num_edges)
         {
            dst->edges = (edge_t *)Q_realloc(dst->edges,sizeof(edge_t) * (dst->num_edges+1));
            if (!dst->edges)
            {
               Q_free(b->faces);
               return NULL;
            }
            dst->edges[dst->num_edges].startvertex = dst->faces[i].verts[j];
            dst->edges[dst->num_edges].endvertex   = dst->faces[i].verts[l];
            dst->num_edges++;
         }
      }
   }
   Q_free(b->faces);

   return dst;
}


int CSG_SplitFace(cface_t *f,plane_t p,cface_t *front,cface_t *back,int keepback)
/* 
  
  Returns:
           0 - Face was clipped successfully
          -1 - Face was completely behind the plane
          -2 - Face was completely in front of the plane
          -3 - Error clipping / culling face 
          -4 - Face was completely ON the plane
*/
{
#define ADDPOINT(f,v) do { if ((f).num_pts==MAX_FACE_POINTS) return -1; (f).pts[(f).num_pts++]=(v); } while (0)

   cface_t FrontFace;
   cface_t BackFace;
   vec3_t p1,p2,p3;
   int i,j;
   double dot;
   int backtest=0;
   int fronttest=0;
   int ontest=0;
   int splitnum=0;
   int sides[MAX_FACE_POINTS];
   double dists[MAX_FACE_POINTS];
   
   if (f->num_pts>MAX_FACE_POINTS) return -3;

   /* determine sides and distances for each point*/
   for (i=0;i<f->num_pts;i++)
   {
      dot = DotProd(f->pts[i], p.n);
      dot -= p.dist;
      dists[i] = dot;
      if (dot > 0.01)
      {
         sides[i] = 1;
         fronttest=1;
      }
      else
      if (dot < -0.01)
      {
         sides[i] = -1;
         backtest=1;
      }
      else
      {
         dists[i]=0;
         sides[i] = 0;
         ontest=1;
      }
   }
   
   if (ontest && !backtest && !fronttest)
      return -4;

   if (backtest^fronttest)
   {
      /*if all in front*/
      if (fronttest)
      {
         *front = *f;
         return -2;
      }
      /*if all in back*/
      if (backtest)
      {
         if (keepback)
            *back = *f;
         return -1;
      }
   }

   FrontFace.num_pts = 0;

   if (keepback)
      BackFace.num_pts = 0;

   for (i=0;i<f->num_pts;i++)
   {
      p1 = f->pts[i];

      if (i==f->num_pts-1)
         j=0;
      else
         j=i+1;

      if (sides[i] == 0)
      {
         ADDPOINT(FrontFace,p1);
         if (keepback)
            ADDPOINT(BackFace,p1);
            
         if ((splitnum==2)||(sides[j]==0))
            return -3;
         splitnum++;
         continue;
      }
   
      if (sides[i] == 1)
         ADDPOINT(FrontFace,p1);
         
      if (sides[i] == -1)
         if (keepback)
            ADDPOINT(BackFace,p1);
      
      if ((sides[j]==sides[i])||(sides[j]==0)) continue;

      /* generate a split point*/
      p2 = f->pts[j];
      
      dot = dists[i] / (dists[i]-dists[j]);
      
      if (p.n.x == 1)
         p3.x = p.dist;
      else
      if (p.n.x == -1)
         p3.x = -p.dist;
      else
         p3.x = p1.x + dot*(p2.x-p1.x);
      
      if (p.n.y == 1)
         p3.y = p.dist;
      else
      if (p.n.y == -1)
         p3.y = -p.dist;
      else
         p3.y = p1.y + dot*(p2.y-p1.y);
      
      if (p.n.z == 1)
         p3.z = p.dist;
      else
      if (p.n.z == -1)
         p3.z = -p.dist;
      else
         p3.z = p1.z + dot*(p2.z-p1.z);
      
      ADDPOINT(FrontFace,p3);
      if (keepback)
         ADDPOINT(BackFace,p3);
      if ((splitnum==2)||(sides[j]==0))
         return -3;
      splitnum++;
   }
      
   if (splitnum!=2)
      return -3;
   front->num_pts = FrontFace.num_pts;
   for (i=0;i<front->num_pts;i++) front->pts[i] = FrontFace.pts[i];
   front->p.n = f->p.n;
   front->p.dist = f->p.dist;

   front->tex = f->tex;
   front->ba = f->ba;

   if (keepback)
   {
      back->num_pts = BackFace.num_pts;
      for (i=0;i<back->num_pts;i++) back->pts[i] = BackFace.pts[i];
      back->p.n = f->p.n;
      back->p.dist = f->p.dist;

      back->tex = f->tex;
      back->tex.g=NULL;
      G_Texdef_Copy(&back->tex,&f->tex);
      back->ba = f->ba;
   }

   return 0;
#undef ADDPOINT
}


int CSG_BrushSide(cbrush_t *b,plane_t p)
{
   int side,bside;
   double d;
   int i,j;

   bside=-2;
   for (i=0;i<b->num_faces;i++)
   {
      for (j=0;j<b->faces[i].num_pts;j++)
      {
         d=DotProd(p.n,b->faces[i].pts[j])-p.dist;
         if (d<-0.01) side=-1;
         else if (d>0.01) side=1;
         else side=0;

         if (!side)
            continue;
            
         if (bside==-2)
            bside=side;
         else
         if (side!=bside)
            return 0;
      }
   }
   return bside;
}


int CSG_BrushesIntersect(cbrush_t *b1,cbrush_t *b2)
{
   int i;

   for (i=0;i<b1->num_faces;i++)
      if (CSG_BrushSide(b2,b1->faces[i].p)==1)
         return 0;
         
   for (i=0;i<b2->num_faces;i++)
      if (CSG_BrushSide(b1,b2->faces[i].p)==1)
         return 0;
   
   return 1;
}


int CSG_SplitBrushWithFace(cbrush_t *Bfront, cface_t p,cbrush_t *Bback)
{                               
   cbrush_t Fbrush,Bbrush;
   cface_t Face1,Face2;
   cface_t NewFace;
   plane_t CutPlane;
   int j,k;
   int fronttest=0;
   int backtest=0;

   switch (CSG_BrushSide(Bfront,p.p))
   {
   case  1: return -1;
   case -1: return -2;
   case -2: return -3;
   }
      
   Fbrush.num_faces = 0;
   Bbrush.num_faces = 0;
   Fbrush.faces=Q_malloc(sizeof(cface_t)*(Bfront->num_faces+1));
   Bbrush.faces=Q_malloc(sizeof(cface_t)*(Bfront->num_faces+1));
   if (!Fbrush.faces || !Bbrush.faces)
      return -3;
      
   for (j=0;j<Bfront->num_faces;j++)
   {
      k = CSG_SplitFace(&Bfront->faces[j],p.p,&Face1,&Face2,TRUE);
      if (k==-4 || k==-3)
         return -3;
      if (k==0)
      {
      /* Face is split by plane, put split sides on corresponding brush */
         Bbrush.faces[Bbrush.num_faces++] = Face2;
         Fbrush.faces[Fbrush.num_faces++] = Face1;
         backtest=fronttest=1;
         continue;
      }
      if (k==-1)
      {
      /* Face is behind plane, put on back brush */
         Bbrush.faces[Bbrush.num_faces++] = Face2;
         backtest=1;
         continue;
      }
      if (k==-2)
      {
      /* Face is in front of plane, put on front brush */
         Fbrush.faces[Fbrush.num_faces++] = Face1;
         fronttest=1;
         continue;
      }
   }
   
   if (!fronttest || !backtest) return -3;

   /* Cap the new "front brush" */
   NewFace.ba = p.ba;
   NewFace.tex = p.tex;
   NewFace.tex.g = NULL;
   G_Texdef_Copy(&NewFace.tex,&p.tex);

   VEC_FlipPlane(NewFace.p, p.p);
   CSG_MakeBoxOnPlane(&NewFace);

   for (j=0;j<Fbrush.num_faces;j++)
   {
      VEC_FlipPlane(CutPlane,Fbrush.faces[j].p);
      k = CSG_SplitFace(&NewFace,CutPlane,&Face1,NULL,FALSE);
      if (k==-4 || k==-3 || k==-1)
         return -3;
      if (k==0)
         NewFace = Face1;
   }
   Fbrush.faces[Fbrush.num_faces++] = NewFace;
   
   /* Cap the new "back brush" */
   NewFace.ba = p.ba;
   NewFace.tex = p.tex;
   NewFace.tex.g = NULL;
   G_Texdef_Copy(&NewFace.tex,&p.tex);
   
   NewFace.p=p.p;
   CSG_MakeBoxOnPlane(&NewFace);

   for (j=0;j<Bbrush.num_faces;j++)
   {
      VEC_FlipPlane(CutPlane,Bbrush.faces[j].p);

      k = CSG_SplitFace(&NewFace,CutPlane,&Face1,NULL,FALSE);
      if (k==-4 || k==-3 || k==-1)
         return -3;
      if (k==0)
         NewFace = Face1;
   }
   Bbrush.faces[Bbrush.num_faces++] = NewFace;
   
   Q_free(Bfront->faces);

   {
      cface_t *f;

      f=Q_realloc(Fbrush.faces,sizeof(cface_t)*Fbrush.num_faces);
      Bfront->num_faces = Fbrush.num_faces;
      Bfront->faces = f?f:Fbrush.faces;

      f=Q_realloc(Bbrush.faces,sizeof(cface_t)*Bbrush.num_faces);
      Bback->num_faces = Bbrush.num_faces;
      Bback->faces = f?f:Bbrush.faces;
   }
   return 0;
}

