#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "hook.h"

#include "memory.h"
#include "s_error.h"
#include "s_message.h"


typedef struct
{
   union
   {
      hook_func_void_t v;
      hook_func_int_t i;
   } f;
   void *data;
   struct plugin_s *p;
   int type;
} hook_func_t;

struct hook_s
{
   char *name;
   int type;
   hook_func_t *hf;
   int num_hf;
};


static hook_t **hooks;
static int num_hooks;


hook_t *Hook_Define(const char *name,int type)
{
   hook_t *h;

   h=Hook_Lookup(name);
   if (h)
   {
      if (h->type!=type)
         S_NewMessage("Warning: Hook %s redefined with type %i!",name,type);
      return h;
   }

   h=Q_malloc(sizeof(hook_t));
   if (!h) S_Abort("Hook_Define","Out of memory!");
   memset(h,0,sizeof(hook_t));

   h->name=Q_strdup(name);
   if (!h->name) S_Abort("Hook_Define","Out of memory!");
   h->type=type;
   hooks=Q_realloc(hooks,sizeof(hook_t *)*(num_hooks+1));
   if (!hooks) S_Abort("Hook_Define","Out of memory!");
   hooks[num_hooks++]=h;
   return h;
}

hook_t *Hook_Lookup(const char *name)
{
   int i;
   for (i=0;i<num_hooks;i++)
      if (!strcmp(hooks[i]->name,name))
         return hooks[i];
   return NULL;
}

void Hook_AddFunc(hook_t *h,int type,hook_func_void_t f,void *data,struct plugin_s *p)
{
   hook_func_t *hf;
   h->hf=Q_realloc(h->hf,sizeof(hook_func_t)*(h->num_hf+1));
   if (!h->hf) S_Abort("Hook_AddFunc","Out of memory!");
   hf=&h->hf[h->num_hf++];
   hf->f.v=f;
   hf->data=data;
   hf->p=p;
   hf->type=type;
}

void Hook_AddFuncLast(hook_t *h,int type,hook_func_void_t f,void *data,struct plugin_s *p)
{
   int i;
   hook_func_t *hf;
   h->hf=Q_realloc(h->hf,sizeof(hook_func_t)*(h->num_hf+1));
   if (!h->hf) S_Abort("Hook_AddFunc","Out of memory!");
   for (i=h->num_hf;i>0;i--)
      h->hf[i]=h->hf[i-1];
   h->num_hf++;

   hf=&h->hf[0];
   hf->f.v=f;
   hf->data=data;
   hf->p=p;
   hf->type=type;
}

int Hook_RemoveFunc(hook_t *h,hook_func_void_t f)
{
   int i;
   int delta;

   delta=0;
   for (i=0;i<h->num_hf;i++)
   {
      if (h->hf[i].f.v!=f)
      {
         if (delta)
            h->hf[i-delta]=h->hf[i];
         continue;
      }
      delta++;
   }
   h->num_hf-=delta;
   return delta;
}

int Hook_RemoveFuncData(hook_t *h,hook_func_void_t f,void *data)
{
   int i;
   int delta;

   delta=0;
   for (i=0;i<h->num_hf;i++)
   {
      if (h->hf[i].f.v!=f || h->hf[i].data!=data)
      {
         if (delta)
            h->hf[i-delta]=h->hf[i];
         continue;
      }
      delta++;
   }
   h->num_hf-=delta;
   return delta;
}

int Hook_RemovePlugin(hook_t *h,struct plugin_s *p)
{
   int i;
   int delta;

   delta=0;
   for (i=0;i<h->num_hf;i++)
   {
      if (h->hf[i].p!=p)
      {
         if (delta)
            h->hf[i-delta]=h->hf[i];
         continue;
      }
      delta++;
   }
   h->num_hf-=delta;
   return delta;
}

int Hook_Call(hook_t *h,void *param)
{
   int i,j;
   if (h->type==HOOK_CALL_HANDLE)
   {
      for (i=h->num_hf-1;i>=0;i--)
      {
         if (h->hf[i].type==HOOK_FUNC_VOID)
         {
            h->hf[i].f.v(param,h->hf[i].data);
            return 1;
         }
         j=h->hf[i].f.i(param,h->hf[i].data);
         if (j) return j;
      }
      return 0;
   }
   for (i=h->num_hf-1;i>=0;i--)
      h->hf[i].f.v(param,h->hf[i].data);
   return 0;
}


static void RemoveHooks(struct plugin_s *p)
{
   int i;

   for (i=0;i<num_hooks;i++)
      Hook_RemovePlugin(hooks[i],p);
}


hook_t
   *hk_plugin_remove,
   *hk_cmd_start,
   *hk_cmd_end,
   *hk_map_new,
   *hk_map_unload,
   *hk_map_load,
   *hk_map_save,
   *hk_map_saved;


void Hook_Init(void)
{
   hooks=NULL;
   num_hooks=0;

   hk_plugin_remove=Hook_Define("plugin.remove",HOOK_CALL_ALL);
   hk_cmd_start=Hook_Define("cmd.start",HOOK_CALL_HANDLE);
   hk_cmd_end=Hook_Define("cmd.end",HOOK_CALL_ALL);

   hk_map_new=Hook_Define("map.new",HOOK_CALL_ALL);
   hk_map_unload=Hook_Define("map.unload",HOOK_CALL_ALL);
   hk_map_load=Hook_Define("map.load",HOOK_CALL_ALL);
   hk_map_save=Hook_Define("map.save",HOOK_CALL_ALL);
   hk_map_saved=Hook_Define("map.saved",HOOK_CALL_ALL);
   
   Hook_AddFunc(hk_plugin_remove,
      HOOK_FUNC_VOID,(hook_func_void_t)RemoveHooks,NULL,
      NULL);
}

