/*
quest.c file of the Quest Source Code

Copyright 1997, 1998, 1999, 2000 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include "defines.h"
#include "types.h"

#include "quest.h"
#include "s_quest.h"

#include "camera.h"
#include "config.h"
#include "hook.h"
#include "map.h"
#include "plugin.h"
#include "s_3d.h"
#include "s_error.h"
#include "s_message.h"
#include "status.h"
#include "undo.h"
#include "version.h"



/* set to zero to see timings */

#if (1)

#define Time(x) x
#define TimeR(x) x

#else

#include "s_sys.h"

#define TimeR(x) \
({ \
   float start; \
   float end; \
   typeof(x) r; \
    \
   start=Sys_Time(); \
   r=x; \
   end=Sys_Time(); \
   printf("--- start=%-5.3f end=%-5.3f elap=%-5.3f %s ---\n",start,end,end-start,#x); \
   r; \
})

#define Time(x) \
({ \
   float start; \
   float end; \
    \
   start=Sys_Time(); \
   x; \
   end=Sys_Time(); \
   printf("--- start=%-5.3f end=%-5.3f elap=%-5.3f %s ---\n",start,end,end-start,#x); \
})

#endif


map_t M; /* Current map. Not a pointer for speed reasons. */


static char **argv;
static int argc;


static char loadfilename[MAX_PATH];


/* Argument parsing */
/* TODO: clean this up. dangerous to use exit() here? should use system-specific thing? */
typedef struct
{
   const char *name;  /* long name (--help) */
   const char *sh;    /* one letter (-h) or NULL */
   const char *desc;  /* description */
   int pass;          /* what pass to call this in */

   int action;  /* what to do */
   union
   {
      void (*f)(void);
      int *i;
      char *s;
   } x;
} arg_t;

/* actions */
#define ACT_CALL   0     /* call the function void data(void) */
#define ACT_SET1   1     /* *data=1 */
#define ACT_SETSTR 2     /* strcpy(data,next_param) */

/* passes */
#define P_0      0  /* the first pass, before anything else */
#define P_LAST 100  /* the last pass, just before returning from InitQuest */

static void ARG_Help(void);

static arg_t args[]=
{
{"help"     ,"h" ,"Get help on arguments" ,P_0   ,ACT_CALL  ,{f:ARG_Help}},

{0          ,   0,                       0,     0,         0,{0}}
};
#define NUM_ARGS (sizeof(args)/sizeof(args[0])-1)

static void ARG_Help(void)
{
   int i;
   int l;

   printf("Syntax:\n");
   printf("  %s", argv[0]);
   for (i=0;i<NUM_ARGS;i++)
   {
      printf(" [--%s]",args[i].name);
   }
   printf("\n\n"
          "Name:          Description:\n");
   for (i=0;i<NUM_ARGS;i++)
   {
      l=0;
      l+=printf("--%s",args[i].name);
      if (args[i].sh)
         l+=printf(" (-%s)",args[i].sh);
      for (;l<15;l++)
         printf(" ");
      printf("%s\n",args[i].desc);
   }
   exit(0);
}

static int curarg;

static void ExecArg(arg_t *a)
{
   switch (a->action)
   {
   case ACT_CALL:
      a->x.f();
      break;
   case ACT_SET1:
      *a->x.i=1;
      break;
   case ACT_SETSTR:
      curarg++;
      if (curarg>=argc)
      {
         printf("Error in argument '%s'! Try '%s --help' for help.\n",
            argv[curarg], argv[0]);
         exit(1);
      }
      strcpy(a->x.s,argv[curarg]);
      break;
   }
}

static void ParseArgs(int pass)
{
   int j;
   char *a;
   int cmap;

   cmap=0;
   for (curarg=1;curarg<argc;curarg++)
   {
      a=argv[curarg];
      if ((a[0]=='-') && (a[1]=='-'))
      {
         a++;
         a++;
         for (j=0;j<NUM_ARGS;j++)
         {
            if (!strcmp(a,args[j].name))
            {
               if (args[j].pass==pass) ExecArg(&args[j]);
               break;
            }
         }
         if (j==NUM_ARGS)
         {
            printf("Unknown argument '%s'! Try '%s --help' for help.\n",
               argv[curarg], argv[0]);
            exit(1);
         }
      }
      else
      if ((a[0]=='-'))
      {
         a++;
         for (j=0;j<NUM_ARGS;j++)
         {
            if (!strcmp(a,args[j].sh))
            {
               if (args[j].pass==pass) ExecArg(&args[j]);
               break;
            }
         }
         if (j==NUM_ARGS)
         {
            printf("Unknown argument '%s'! Try '%s --help' for help.\n",
               argv[curarg], argv[0]);
            exit(1);
         }
      }
      else
      {
         if (pass==P_0)
            strcpy(loadfilename,argv[curarg]);
      }
   }
}


void InitQuestCore(void)
{
   Time(Status_Init());
   Time(Hook_Init());
}


void InitQuest(int pargc, char **pargv)
{
#ifndef ALPHA
   printf("\n"
          "\n"
          "The \"Official\" new DOS/Win95 Quest " QUEST_VERSION " (" __DATE__ ")\n"
          "-----------------------------------------------------------------\n"
          " (c) Copyright 1997, 1998, 1999, 2000\n"
          "     Alexander Malmberg <alexander@malmberg.org>\n"
          "\n"
          " (c) Copyright 1996,\n"
          "     Chris Carollo\n"
          "     Trey Harrison\n"
          "\n"
          "Please read all of the documents contained in the Quest zipfile. They\n"
          "contain useful information.\n"
          "\n"
          "This program is distributed under the terms of the GNU General Public\n"
          "License as published by the Free Software Foundation. A copy of it is in\n"
          "the file legal.txt. This program comes with ABSOLUTELY NO WARRANTY.\n"
          "\n");
#ifdef BETA
   printf("This is a beta version of Quest. If you are not an official Quest beta\n"
          "tester, you are encouraged to use the latest stable version instead.\n"
          "\n");
#endif
#endif /* ALPHA */

   argv=pargv;
   argc=pargc;

   /* -------------- */
   /* INITIALIZATION */
   /* -------------- */
   memset(&M,0,sizeof(M)); /* this should zero out ALL map stuff */

   Time(ParseArgs(P_0));

   Time(Undo_Init());

   Time(Plugin_Init());

   if (loadfilename[0])
   {
      char *d;
      int ext;

      ext=1;
      for (d=loadfilename;*d;d++)
      {
         if (*d=='.')
            ext=0;
         if ((*d=='/') || (*d=='\\'))
            ext=1;
      }
   
      if (ext)
         strcat(loadfilename,".map");

      if (!TimeR(Map_Load(loadfilename)))
      {
         S_Error("InitQuest", "Unable to load %s.",loadfilename);
         Map_New();
      }
   }
   else
      Time(Map_New());
   
   /* Update all viewports */
   Time(S_UpdateAllViewports());
   
   ParseArgs(P_LAST);
   
   S_NewMessage("Quest " QUEST_VERSION " initialized.");

   Undo_Done();
}


void DoneQuest(void)
{
}

