/* TODO: should be cmd/edit/trans.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "camera.h"
#include "error.h"
#include "map.h"
#include "matrix.h"
#include "quest.h"
#include "sel.h"
#include "transform.h"
#include "undo.h"

#include "fe/3d.h"
#include "fe/input.h"
#include "fe/int.h"
#include "fe/message.h"
#include "fe/trans.h"


/* TODO: configurable? */
#define STEP 8

#define INSTRUCTIONS \
	"Left-click when done, right-click to cancel. Hold down shift to move " \
	"mouse without changes. Shift-click to place focal point (2d viewports " \
	"only). Shift-right-click to change axis."

/* TODO: can't have key name in function descriptions */
static const char tdoc[]=
"The edit.trans.* function are used to do different linear transforms "
"(rotations, scalings, mirrorings) on objects.\n"
"\n"
"To use them, simply select the objects you want to transform and hit "
"ctrl-r to rotate, ctrl-m to mirror, or ctrl-g to scale. Moving the "
"mouse around will rotate/mirror/scale the selected objects. If you "
"want to move the mouse without affecting the objects, hold shift down. "
"To change the axis the transform uses, shift-right-click. In 2d viewports, "
"you can select which point to rotate/scale/mirror around by shift-clicking "
"in the viewport. The default point is the center of the selection.\n"
"\n"
"When you're done, left-click to place the objects. You can also right-click "
"to cancel the command.\n"
"\n"
"The commands will transform all selected vertices. If no vertices are "
"selected, all vertices on all selected brushes and faces will be "
"transformed. Additionally, any entities that are selected will be "
"transformed, and if an entity has any brushes attached, they will also be "
"transformed.";


static void transform(const char *msg,
	int max_axis,int axis,char const **str_axis,
	void (*get_matrix)(int axis,int d,matrix_t m))
{
	s_event_t e;
	sel_t new;

	matrix_t m,m2;

	int iscale,oscale;
	vec3_t focal,center;

	int vaxis;


	{
		vec3_t d;
		Move90(S_cur_vport,MOVE_FORWARD,&d,16);
		     if (fabs(d.x)>10) vaxis=0;
		else if (fabs(d.y)>10) vaxis=1;
		else if (fabs(d.z)>15) vaxis=2; /* TODO: why 15? */
		else vaxis=-1;
	}

	if (!Trans_Setup(&M.sel,&new))
		return;

	Trans_Center(&new,&center);
	focal=center;
	SnapVecToGrid(&focal);

	S_I_Setup(&e,S_I_MOTION|S_I_CLICK,msg);

	iscale=0;
	oscale=-1;
	while (1)
	{
		if (oscale!=iscale/STEP)
		{
			oscale=iscale/STEP;

			InitMatrix(m);
			m[0][3]=focal.x;
			m[1][3]=focal.y;
			m[2][3]=focal.z;

			get_matrix(axis,oscale,m2);

			MultMatrix(m,m2,m2);

			InitMatrix(m);
			m[0][3]=-focal.x;
			m[1][3]=-focal.y;
			m[2][3]=-focal.z;

			MultMatrix(m2,m,m);

			Trans_Linear(m,&new);
			Undo_Update();

			S_UpdateAllViewports();
		}

		S_I_Wait(&e);
		if (e.type==S_I_CLICK)
		{
			if (e.key==(KEYM_SHIFT|KEY_BUTTON2))
			{
				axis++;
				if (axis==max_axis) axis=0;
				S_NewMessage("%s",str_axis[axis]);
				oscale=1;
				iscale=0;
			}
			else
			if (e.key==(KEYM_SHIFT|KEY_BUTTON1))
			{
				vec3_t m;
				if (vaxis!=-1 && S_R_GetMousePos(&m) && S_R_Trans_SVert(m,&focal))
				{
					switch (vaxis)
					{
					case 0: focal.x=center.x; break;
					case 1: focal.y=center.y; break;
					case 2: focal.z=center.z; break;
					}
					SnapVecToGrid(&focal);
					oscale=1;
					iscale=0;
				}
			}
			else
				break;
		}
		if (e.type==S_I_MOTION && !(e.key&KEYM_SHIFT))
			iscale+=e.dx+e.dy;
	}

	if ((e.key&KEY_MASK)==KEY_BUTTON1)
	{
		S_NewMessage("Operation completed.");
		Trans_Done(&new,0);
	}
	else
	{
		S_NewMessage("Operation canceled.");
		Trans_Done(&new,1);
	}
	S_I_Done();
}


static void scale_matrix(int axis,int d,matrix_t m)
{
	InitMatrix(m);
	switch (axis)
	{
	case 0: m[0][0]=1+d*0.1; break;
	case 1: m[1][1]=1+d*0.1; break;
	case 2: m[2][2]=1+d*0.1; break;
	case 3: m[0][0]=m[1][1]=m[2][2]=1+d*0.1; break;
	default: ASSERT(1);
	}
}

static void C_edit_trans_scale(void)
{
static const char *saxis[]=
	{"X-axis","Y-axis","Z-axis","All axises"};
	transform(
		"Move the mouse to scale the selection. " INSTRUCTIONS,
		4,3,saxis,scale_matrix);
}
static func_t F_edit_trans_scale=
{"edit.trans.scale",FUNC_VOID,{fv:C_edit_trans_scale},NULL,NULL,0,
{"Scale the selected objects.",tdoc}
};


static void rotate_matrix(int axis,int d,matrix_t m)
{
	double s,c;
	double ang=d*5;

	ang=ang/180.0*PI;
	s=sin(ang);
	c=cos(ang);

	InitMatrix(m);
	switch (axis)
	{
	case 0:
		m[1][1]= c;
		m[1][2]= s;

		m[2][1]=-s;
		m[2][2]= c;
		break;

	case 1:
		m[0][0]= c;
		m[0][2]= s;

		m[2][0]=-s;
		m[2][2]= c;
		break;

	case 2:
		m[0][0]= c;
		m[0][1]= s;

		m[1][0]=-s;
		m[1][1]= c;
		break;

	default: ASSERT(1);
	}
}

static void C_edit_trans_rotate(void)
{
static const char *saxis[]=
	{"X-axis","Y-axis","Z-axis"};
	int i;
	vec3_t d;

	Move90(S_cur_vport,MOVE_FORWARD,&d,16);
	     if (fabs(d.x)>10) i=0;
	else if (fabs(d.y)>10) i=1;
	else                   i=2;

	transform(
	   "Move the mouse to rotate the selection. " INSTRUCTIONS,
	   3,i,saxis,rotate_matrix);
}
static func_t F_edit_trans_rotate=
{"edit.trans.rotate",FUNC_VOID,{fv:C_edit_trans_rotate},NULL,NULL,0,
{"Rotate the selected objects.",tdoc}
};


static void mirror_matrix(int axis,int d,matrix_t m)
{
	InitMatrix(m);
	if (d>0) return;
	switch (axis)
	{
	case 0: m[0][0]=-1; break;
	case 1: m[1][1]=-1; break;
	case 2: m[2][2]=-1; break;
	default: ASSERT(1);
	}
}

static void C_edit_trans_mirror(void)
{
static const char *saxis[]=
	{"X-axis","Y-axis","Z-axis"};

	vec3_t d;
	int i;
	
	Move90(S_cur_vport,MOVE_LEFT,&d,16);
	     if (fabs(d.x)>10) i=0;
	else if (fabs(d.y)>10) i=1;
	else                   i=2;

	transform(
	   "Move the mouse to mirror the selection. " INSTRUCTIONS,
	   3,i,saxis,mirror_matrix);
}
static func_t F_edit_trans_mirror=
{"edit.trans.mirror",FUNC_VOID,{fv:C_edit_trans_mirror},NULL,NULL,0,
{"Mirror the selected objects.",tdoc}
};


static func_t *funcs[]={&F_edit_trans_scale,
	&F_edit_trans_mirror,&F_edit_trans_rotate};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"edit.trans","Object transforming commands (rotate, scale, mirror)."}};

PLUGIN_FOOT_FUNCS

