#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>

#include "defines.h"
#include "types.h"

#include "error.h"
#include "err_catch.h"

#include "fe/error.h"
#include "fe/message.h"


error_handler_t *error_handler;


static char *errors;
static int errors_len,errors_size;


void Abort_Str(const char *str)
{
	FE_Abort("Critical error",errors_len?"%s\n%s\n\n":"%s\n\n",str,errors);
}

void Error_Str(const char *str)
{
	int le,ls;
	if (!error_handler)
	{
#if 0
		fprintf(stderr,"*** Unhandled error ***\n%s\n",str);
		if (errors_len) fprintf(stderr,"%s\n",errors);
		fprintf(stderr,"\n");
		abort();
#else
		FE_Abort("Unhandled error",errors_len?"%s\n%s\n\n":"%s\n\n",str,errors);
#endif
	}
	if (errors_len)
		le=errors_len+1;
	else
		le=0;
	ls=strlen(str);
	if (le+ls+1>errors_size)
	{
		char *e;
		e=realloc(errors,le+ls+1);
		if (!e)
		{
#if 0
			fprintf(stderr,"*** Unhandled error ***\n%s\n%s\n",
				"Out of memory while trying to recover from error!",
				str);
			if (errors_len) fprintf(stderr,"%s\n",errors);
			fprintf(stderr,"\n");
			abort();
#else
			FE_Abort("Out of memory while trying to recover from error",
				errors_len?"%s\n%s\n\n":"%s\n\n",str,errors);
#endif
		}
		errors=e;
		errors_size=le+ls+1;
	}
	if (le)
	{
		memmove(&errors[ls+1],errors,le);
		strcpy(errors,str);
		errors[ls]='\n';
	}
	else
		strcpy(errors,str);
	errors_len+=ls+1;

	{
		error_handler_t *cur=error_handler;
		error_handler=cur->next;
		longjmp(cur->j,1);
	}
}

void Error_Add_Str(const char *str)
{
	int le,ls;
	if (errors_len)
		le=errors_len+1;
	else
		le=0;
	ls=strlen(str);
	if (le+ls+1>errors_size)
	{
		char *e;
		e=realloc(errors,le+ls+1);
		if (!e)
		{
			FE_Abort("Out of memory while trying to recover from error",
				errors_len?"%s\n%s\n\n":"%s\n\n",str,errors);
		}
		errors=e;
		errors_size=le+ls+1;
	}
	if (le)
	{
		memmove(&errors[ls+1],errors,le);
		strcpy(errors,str);
		errors[ls]='\n';
	}
	else
		strcpy(errors,str);
	errors_len+=ls+1;
}

void Warning_Str(const char *str)
{
	S_NewMessage("Warning: %s",str);
}


void Abort(const char *format, ...)
{
	char buf[1024];
	va_list args;

	va_start(args,format);
	vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	Abort_Str(buf);
}

void Error(const char *format, ...)
{
	char buf[1024];
	va_list args;

	va_start(args,format);
	vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	Error_Str(buf);
}

void Error_Add(const char *format, ...)
{
	char buf[1024];
	va_list args;

	va_start(args,format);
	vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	Error_Add_Str(buf);
}

void Warning(const char *format, ...)
{
	char buf[1024];
	va_list args;

	va_start(args,format);
	vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	Warning_Str(buf);
}


void Abort_File(const char *file,const char *function,int line,const char *format, ...)
{
	char buf[1024];
	va_list args;
	int l;

	va_start(args,format);
	l=vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	buf[l]=' ';
	snprintf(&buf[l+1],sizeof(buf)-l-1,"(%s:%s:%i)",file,function,line);
	Abort_Str(buf);
}

void Error_File(const char *file,const char *function,int line,const char *format, ...)
{
	char buf[1024];
	va_list args;
	int l;

	va_start(args,format);
	l=vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	buf[l]=' ';
	snprintf(&buf[l+1],sizeof(buf)-l-1,"(%s:%s:%i)",file,function,line);
	Error_Str(buf);
}

void Error_Add_File(const char *file,const char *function,int line,const char *format, ...)
{
	char buf[1024];
	va_list args;
	int l;

	va_start(args,format);
	l=vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	buf[l]=' ';
	snprintf(&buf[l+1],sizeof(buf)-l-1,"(%s:%s:%i)",file,function,line);
	Error_Add_Str(buf);
}

void Warning_File(const char *file,const char *function,int line,const char *format, ...)
{
	char buf[1024];
	va_list args;
	int l;

	va_start(args,format);
	l=vsnprintf(buf,sizeof(buf),format,args);
	va_end(args);
	buf[l]=' ';
	snprintf(&buf[l+1],sizeof(buf)-l-1,"(%s:%s:%i)",file,function,line);
	Warning_Str(buf);
}


void Error_Clear(void)
{
	errors_len=0;
}

void Error_Display(const char *format, ...)
{
	if (format)
	{
		char buf[1024];
		va_list args;

		va_start(args,format);
		vsnprintf(buf,sizeof(buf),format,args);
		va_end(args);
		Error_Add(buf);
	}
	FE_Error(errors);
	Error_Clear();
}

int Error_Have(void)
{
	return errors_len>0;
}

