/*
undo.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "undo.h"

#include "brush.h"
#include "edsel.h"
#include "entity.h"
#include "error.h"
#include "group.h"
#include "map.h"
#include "memory.h"
#include "ql.h"
#include "quest.h"
#include "uid.h"
#include "fe/message.h"
#include "fe/render.h"


#include "config.h"


static config_var_t
	*cv_undo_min_levels,
	*cv_undo_max_levels,
	*cv_undo_max_mem,
	*cv_undo_combine;


/* TODO: Handle out-of-memory-errors better. If we run out of memory
while saving undo information, we should safely be able to call Cmd_Abort
and use the partial undo information to restore things, at least as long
as we're careful about only linking in valid stuff.
*/

/*

When saving undo information:

              Old type
             X            NEW          MOD           DEL
New     NEW  add as NEW   error        error         error
type    MOD  add as MOD   do nothing   do nothing    error
        DEL  add as DEL   remove       mark as DEL   error

When undoing:
Type
NEW   delete original
MOD   delete original, replace with copy
DEL   add copy

*/


#define T_NEW    0 /* uses only u field */
#define T_MODIFY 1
#define T_DELETE 2
typedef struct undo_object_s
{
	struct undo_object_s *Next,*Last;
	int type;
	map_uid_t u;
	union
	{
		brush_t b;
		entity_t e;
		group_t g;
	} d;
} undo_object_t;


typedef struct undo_info_s
{
	struct undo_info_s *Next,*Last;
	char *name;
	int combine;

	undo_object_t *b;
	undo_object_t *e;
	undo_object_t *g;
} undo_info_t;


typedef struct undo_s
{
	undo_info_t *UndoHead;
} undo_t;


static undo_info_t *CurUndo;

static int cur_flags;


static int Undo_Size(undo_info_t *u)
{
	undo_object_t *o;
	int s;

	s=0;
	for (o=u->b;o;o=o->Next)
		s+=sizeof(undo_object_t)+B_Size(&o->d.b);
	for (o=u->e;o;o=o->Next)
		s+=sizeof(undo_object_t)+E_Size(&o->d.e);
	for (o=u->g;o;o=o->Next)
		s+=sizeof(undo_object_t)+G_Size(&o->d.g);

	s+=strlen(u->name)+1;
	s+=sizeof(undo_info_t);

	return s;
}


static void Undo_Free(undo_info_t *u)
{
	undo_object_t *o;

	if (u->name) Q_free(u->name);

	for (o=u->b;o;o=u->b)
	{
		if (o->type!=T_NEW)
			B_FreeC(&o->d.b);
		u->b=o->Next;
		Q_free(o);
	}

	for (o=u->e;o;o=u->e)
	{
		if (o->type!=T_NEW)
			E_FreeC(&o->d.e);
		u->e=o->Next;
		Q_free(o);
	}

	for (o=u->g;o;o=u->g)
	{
		if (o->type!=T_NEW)
			G_FreeC(&o->d.g);
		u->g=o->Next;
		Q_free(o);
	}

	if (u->Next) u->Next->Last=u->Last;
	if (u->Last) u->Last->Next=u->Next;

	if (u==M.undo->UndoHead) M.undo->UndoHead=u->Next;
	Q_free(u);
}


static undo_object_t *FindObject(undo_object_t *s,map_uid_t u)
{
	for (;s;s=s->Next)
		if (s->u==u) return s;
	return NULL;
}

static undo_object_t *NewObject(undo_object_t **l)
{
	undo_object_t *o;

	o=Q_malloc(sizeof(undo_object_t));
	if (!o) ABORT("Undo.NewObject: Out of memory!");
	memset(o,0,sizeof(undo_object_t));

	o->Next=*l;
	if (*l) (*l)->Last=o;
	*l=o;
	return o;
}

static void DeleteObject(undo_object_t *o,undo_object_t **l)
{
	if (o->Next) o->Next->Last=o->Last;
	if (o->Last) o->Last->Next=o->Next;
	if (o==*l) *l=o->Next;
	Q_free(o);
}


/* Lots of utility functions call these, and they might also
work on non-map data, so we check x->real here. If it isn't
set, we just ignore the call. */
#define UNDP(t,n,l,L) \
void U_##n##Add(t *l)                              \
{                                                  \
	undo_object_t *o;                               \
	                                                \
	if (cur_flags==2) return;                       \
	if (!l->real) return;                           \
	ASSERT(cur_flags!=1);                           \
	if (!CurUndo) Undo_NewDefault();                \
	                                                \
	o=FindObject(CurUndo->l,l->uid);                \
	ASSERT(!o);                                     \
	o=NewObject(&CurUndo->l);                       \
	o->type=T_NEW;                                  \
	o->u=l->uid;                                    \
}                                                  \
                                                   \
void U_##n##Modify(t *l)                           \
{                                                  \
	undo_object_t *o;                               \
	                                                \
	if (cur_flags==2) return;                       \
	if (!l->real) return;                           \
	ASSERT(cur_flags!=1);                           \
	if (!CurUndo) Undo_NewDefault();                \
	                                                \
	o=FindObject(CurUndo->l,l->uid);                \
	if (!o)                                         \
	{                                               \
		o=NewObject(&CurUndo->l);                    \
		o->type=T_MODIFY;                            \
		o->u=l->uid;                                 \
		if (! L##_Duplicate(&o->d.l,l))              \
			ABORT("Out of memory!");                  \
		o->d.l.uid=o->u;                             \
		return;                                      \
	}                                               \
	ASSERT(o->type!=T_DELETE);                      \
}                                                  \
                                                   \
void U_##n##Delete(t *l)                           \
{                                                  \
	undo_object_t *o;                               \
	                                                \
	if (cur_flags==2) return;                       \
	if (!l->real) return;                           \
	ASSERT(cur_flags!=1);                           \
	if (!CurUndo) Undo_NewDefault();                \
	                                                \
	o=FindObject(CurUndo->l,l->uid);                \
	if (!o)                                         \
	{                                               \
		o=NewObject(&CurUndo->l);                    \
		o->type=T_DELETE;                            \
		o->u=l->uid;                                 \
		if (! L##_Duplicate(&o->d.l,l))              \
			ABORT("Out of memory!");                  \
		o->d.l.uid=o->u;                             \
		return;                                      \
	}                                               \
	ASSERT(o->type!=T_DELETE);                      \
	switch (o->type)                                \
	{                                               \
	case T_NEW:                                     \
		DeleteObject(o,&CurUndo->l);                 \
		break;                                       \
	case T_MODIFY:                                  \
		o->type=T_DELETE;                            \
		break;                                       \
	}                                               \
}


static void Undo_NewDefault(void)
{
	Undo_New(NULL,0,0);
}


UNDP(struct brush_s,Brush,b,B)
UNDP(struct entity_s,Entity,e,E)
UNDP(struct group_s,Group,g,G)

#undef UNDP


void Undo_New(const char *name,int combine,int flags)
{
	if (flags)
	{
		cur_flags=flags;
		if (flags==2)
		{
			Undo_Clear();
		}
		return;
	}

	if (!M.undo)
	{
		M.undo=Q_malloc(sizeof(undo_t));
		if (!M.undo) ABORT("Undo.New: Out of memory!");
		memset(M.undo,0,sizeof(undo_t));
	}

	CurUndo=Q_malloc(sizeof(undo_info_t));
	if (!CurUndo) ABORT("Undo.New: Out of memory!");
	memset(CurUndo,0,sizeof(undo_info_t));

	if (name)
	   CurUndo->name=Q_strdup(name);
	else
		CurUndo->name=Q_strdup("default");
	if (!CurUndo->name) ABORT("Undo.New: Out of memory!");
	CurUndo->combine=combine;

	if (M.undo->UndoHead) M.undo->UndoHead->Last=CurUndo;
	CurUndo->Next=M.undo->UndoHead;
	M.undo->UndoHead=CurUndo;
}

void Undo_SetClear(void)
{
	if (cur_flags==2) return;
	Undo_Clear();
	cur_flags=2;
}


static void Undo_Check(void);

/* Combines src and dst into dst, frees src */
static void Undo_Combine(undo_info_t *dst,undo_info_t *src)
{
	undo_info_t *cu;
	undo_object_t *o;

	cu=CurUndo;
	CurUndo=dst;

#define UNDC(n,l)                                     \
	for (o=src->l;o;o=o->Next)                         \
	{                                                  \
		switch (o->type)                                \
		{                                               \
		case T_NEW:    U_##n##Add   (&o->d.l); break;   \
		case T_MODIFY: U_##n##Modify(&o->d.l); break;   \
		case T_DELETE: U_##n##Delete(&o->d.l); break;   \
		}                                               \
	}

	UNDC(Brush,b)
	UNDC(Entity,e)
	UNDC(Group,g)

	Undo_Free(src);
	CurUndo=cu;
}


void Undo_Done(void)
{
	if (cur_flags==2)
	{
		Undo_Clear();
		cur_flags=0;
		return;
	}
	if (cur_flags==1)
	{
		cur_flags=0;
		return;
	}


/* If nothing changed (cancelled command, or commands that don't cause
any changes) we don't want to use up an undo slot. */
	if (!CurUndo) return;
	if (!CurUndo->b && !CurUndo->e && !CurUndo->g)
	{
		Undo_Free(CurUndo);
		CurUndo=NULL;
		return;
	}

	ASSERT(M.undo);


/*
If we can combine it with the last unfo_info_t, do so (e.g. if
we have many move_sel_vert_dir:s after each other, we want to
combine them, both for memory reasons and to make undo more useful).
Actually, we might want to wait with combining and try to combine
CurUndo->Next and CurUndo->Next->Next here instead. That way,
one 'undo' undos the last vertex movement (which is what you want
if the last one was off and degenerated the brush or something)
and two 'undo':s undos the entire movement sequence (which is
probably what you want to do if you don't want to undo just the last
move).
*/
	if (cv_undo_combine->i &&
		 M.undo->UndoHead->Next && M.undo->UndoHead->Next->Next)
	{
		undo_info_t *u1,*u2;
		u1=M.undo->UndoHead->Next;
		u2=u1->Next;
		if (u1->combine && u2->combine &&
			 !strcmp(u1->name,u2->name))
			Undo_Combine(u2,u1);
	}

/* Optimize undo information here?

For many brush operations, the planes won't change at all, but
we're still duplicating them. We could save lots of memory by
comparing 'old' and 'new' structs here and only saving the
differences.
For brushes, we could avoid duplicating:
  faces     useful for: vertex movement
  verts                 texture changes
  edges                 vertex movement, texture changes

For large scale operations, like removing an entity or
creating/removing groups, that affect large amounts of brushes,
this could have a large impact. Things like moving an entire
map around a bit would also be helped a lot.

Problems:
  There'll be a speed hit if we need to compare lots of stuff.
  If something goes wrong (command failure via Cmd_Abort,
  or internal consistency error), we might lose the information
  we optimized away and we might not be able to restore things.

Idea: pass a bitmask of stuff that changes to U_BrushModify,
e.g. UBRUSH_VERTS, UBRUSH_EDGES, UBRUSH_PLANES. When only
changing vertices, you'd do U_BrushModify(b,UBRUSH_VERTS); .
When merging vertices or splitting faces, you'd do a
U_BrushModify(b,UBRUSH_VERTS|UBRUSH_EDGES|UBRUSH_PLANES); .
Texture changes would do U_BrushModify(b,UBRUSH_PLANES); .
Then in this function, we just go over the list of brushes
and remove whatever isn't flagged as necessary.


For entities, we could compare keys/values and not
duplicate them for e.g. group changes, but it
probably isn't worth the effort. For groups,
there's nothing to be done, but they're small
and you usually don't have that many of them.

*/

	Undo_Update();
	M.modified=1;

	CurUndo=NULL;

	Undo_Check();
}


void Undo_Clear(void)
{
	undo_info_t *u;

	if (!M.undo) return;

	for (u=M.undo->UndoHead;u;u=M.undo->UndoHead)
		Undo_Free(u);
	Q_free(M.undo);
	M.undo=NULL;
}



static void Undo_Check(void)
{
	undo_info_t *o,*on;
	int tsize,levels;

	tsize=levels=0;
	for (o=on=M.undo->UndoHead;on;on=on->Next)
	{
		o=on;
		levels++;
		tsize+=Undo_Size(o);
	}

	while (o && levels>cv_undo_min_levels->i &&
			 ((cv_undo_max_levels->i && levels>cv_undo_max_levels->i) ||
			  (cv_undo_max_mem->i && tsize>cv_undo_max_mem->i*1024)) )
	{
		on=o->Last;
		levels--;
		tsize-=Undo_Size(o);
		Undo_Free(o);
		o=on;
	}
}


void Undo_Update(void)
{
	undo_object_t *o;
	brush_t *b;
	entity_t *e;
	group_t *g;

	if (cur_flags==2)
	{
		Map_Update();
		return;
	}
	if (cur_flags==1)
		return;

	ASSERT(CurUndo);

	if (!CurUndo->b && !CurUndo->e && !CurUndo->g)
		return;

#define UNDU(n,l,L)                   \
	for (o=CurUndo->l;o;o=o->Next)     \
		if (o->type!=T_DELETE)          \
			if ((l=UID_##n(o->u)))       \
				L##_Update(l);

	UNDU(Brush,b,B)
	UNDU(Entity,e,E)
	UNDU(Group,g,G)

#undef UNDU

	S_R_SetUpdate(R_UPDATE_ALL);

	Map_UpdateVLists();
}


static void Undo(void)
{
	undo_object_t *o;
	undo_info_t *u;
	brush_t *b;
	entity_t *e;
	group_t *g;

	bsel_t *bs;
	esel_t *es;

	u=M.undo->UndoHead;
	ASSERT(u);

#define UNDT(n,ns,l,L,rscall)                              \
	for (o=u->l;o;o=o->Next)                                \
	{                                                       \
		l=NULL;                                              \
		switch (o->type)                                     \
		{                                                    \
		case T_NEW:                                          \
		case T_MODIFY:                                       \
			/* Find original */                               \
			l=UID_##n(o->u);                                  \
			if (!l)                                           \
			{                                                 \
				Warning("Undo: "                               \
	"Internal error: UID_%s(%i) failed for %i!",            \
					#n,o->u,o->type);                           \
			}                                                 \
			else                                              \
			{ /* Delete it */                                 \
				/* Unlink manually since we don't want the     \
				Undo_DeleteXxx to catch this */                \
				M.num_##ns--;                                  \
				if (l->Next != NULL) l->Next->Last = l->Last;  \
				if (l->Last != NULL) l->Last->Next = l->Next;  \
				if (l==M.n##Head) M.n##Head=l->Next;           \
				l->Next=NULL;                                  \
				l->Last=NULL;                                  \
				l->real=0;                                     \
				if (o->type==T_NEW)                            \
				{                                              \
					rscall                                      \
					L##_Free(l);                                \
					break;                                      \
				}                                              \
			}                                                 \
		case T_DELETE:                                       \
			if (o->type==T_DELETE) l=NULL;                    \
			/* Add back our copy */                           \
			l=L##_Duplicate(l,&o->d.l);                       \
			/* Link manually so we keep the uid */            \
			l->uid=o->u;                                      \
			if (M.n##Head!=NULL) M.n##Head->Last = l;         \
			l->Next = M.n##Head;                              \
			l->Last = NULL;                                   \
			M.n##Head = l;                                    \
			l->real=1;                                        \
			M.num_##ns++;                                     \
			break;                                            \
		}                                                    \
	}

	UNDT(Brush,brushes,b,B, if ((bs=FindSelBrush(b))) RemoveSelBrush(bs); )
	UNDT(Entity,entities,e,E, if ((es=FindSelEnt(e))) RemoveSelEnt(es); )
	UNDT(Group,groups,g,G,)

#undef UNDT

	Undo_Free(u);

/* This might be slow, but it's the safest way of doing it, so
it's probably worth it. */
	Map_Update();
}

void Undo_Restore(void)
{
	ASSERT(cur_flags==0);
	CurUndo=NULL;
	Undo();
}

void Undo_Abort(void)
{
	if (cur_flags==0)
	{
		CurUndo=NULL;
		Undo();
		return;
	}
	Warning("Undo_Abort: Unable to restore map data! Map data might be corrupted!");
	if (CurUndo)
	{
		Undo_Free(CurUndo);
		CurUndo=NULL;
	}
	cur_flags=0;
	Map_Update();
}


#include "qfunc.h"


static void C_undo(void)
{
	if (!M.undo || !M.undo->UndoHead)
	{
		Error_Add("undo: No more undo information!");
		return;
	}
	ClearSelAll();
	S_NewMessage("Undoing command '%s'.",M.undo->UndoHead->name);
	Undo();
	S_R_SetUpdate(R_UPDATE_ALL);
}
static func_t F_undo=
{"undo",FUNC_VOID,{fv:C_undo},0,0,0,
{"Undo the last editing operation.",
"This command undos the last editing operation. Quest's undo system tracks "
"all changes to brushes, entities, and groups in the map, and this command "
"uses that information to undo the last change. You can use it repeatedly "
"to undo the last several changes, until you run out of undo information "
"(or changes)."
"\n\n"
"If 'undo.combine' is set, some commands (e.g. moving vertices with the "
"keyboard) are grouped together "
"in the undo information. Thus, after using those commands several times "
"after each other, the first undo will undo the last command, and the "
"second undo will undo the entire sequence."
"\n\n"
"The amount of undo information kept by Quest is controlled by three "
"settings:\n\n"
"'undo.max_levels' is the maximum number of undo levels stored by "
"Quest, or 0 for no maximum.\n\n"
"'undo.max_mem' is the maximum amount of memory spent on undo information, "
"in kb (or 0 for no maximum, but that's not recommended).\n\n"
"'undo.min_levels' is the minimum number of undo levels kept. Quest will "
"_always_ keep at least this many levels, even if it exceeds the other "
"settings."},
};


static void C_undo_stats(void)
{
	undo_object_t *o;
	undo_info_t *u;
	int level;
	int tsize;

	printf("M.undo=%p  cur_flags=%i  CurUndo=%p\n",M.undo,cur_flags,CurUndo);
	if (!M.undo) return;

	level=tsize=0;
	for (u=M.undo->UndoHead;u;u=u->Next)
	{
		printf(" --- Undo %i %p ---\n",level,u);
		printf("  Combine=%i  Name='%s'\n",u->combine,u->name);

		printf("  Brushes: ");
		for (o=u->b;o;o=o->Next) printf("(%i:%i) ",o->type,o->u);
		printf("\n");
		printf("  Entities: ");
		for (o=u->e;o;o=o->Next) printf("(%i:%i) ",o->type,o->u);
		printf("\n");
		printf("  Groups: ");
		for (o=u->g;o;o=o->Next) printf("(%i:%i) ",o->type,o->u);
		printf("\n");

		printf("  Size: %i\n",Undo_Size(u));
		tsize+=Undo_Size(u);
	}
	printf("Total size: %i\n",tsize);
}
static func_t F_undo_stats=
{"undo.stats",FUNC_VOID,{fv:C_undo_stats},NULL,NULL,0};


static func_t *funcs[]={&F_undo,&F_undo_stats};


void Undo_Init(void)
{
	cv_undo_combine=Cfg_DefVar(config_quest,
		"undo.combine","1",
		"Whether to combine undo information for sequences of similar "
		"commands.");
		
	cv_undo_min_levels=Cfg_DefVar(config_quest,
		"undo.min_levels","2",
		"Minimum number of undo levels kept, even in they exceed undo.max_mem. "
		"Set to 0 for no minimum.");
	cv_undo_max_levels=Cfg_DefVar(config_quest,
		"undo.max_levels","50",
		"Maximum number of undo levels kept. Set to 0 for no maximum. "
		"undo.max_mem also applies.");
	cv_undo_max_mem   =Cfg_DefVar(config_quest,
		"undo.max_mem","256",
		"Maximum amount of memory spent on undo information, in kb. "
		"undo.max_levels also applies. If undo.min_levels is set, the maximum "
		"might be exceeded. Set to 0 for no maximum (not recommended).");

	Func_Register(NULL,funcs,sizeof(funcs));

	cur_flags=2;
}

